--  |--------------------------------------------------------------------------------------------------
--  | October 2003
--  |
--  |                    General Bus Functional Model
--  |                          by Nir Hamzani
--  |                    Sital Technology 052-851104
--  |
--  |--------------------------------------------------------------------------------------------------
--  |    The language supported by this CPU BFM is as follows :
--  |
--  | legend:
--  |          [must field]
--  |          <optional field>
--  |
--  |
--  |    Command           Parameters              Explanation
--  |    -------    ----------------------     ----------------------
--  |    TEST       [I:integer] <S:string>     Test number Header. Writes 40 characters of S to screen.
--  |    WAIT_FOR   [T:time]                   causes a delay for time T .
--  |    ECHO       [S:string]                 first 40 characters of S only.
--  |    LOOP_FOR   [I:integer]                Loops from here to END LOOP for I times.
--  |    END_LOOP
--  |    BREAK                                 Stops BFM Simulation.
--  |    READ       [Address] [Data]           Read from address (hex mode generic width) and
--  |                                          compare to the expected data. Data may be - (don't care)
--  |                                          if no data is needed to compare.
--  |
--  |    VERIFY_FILL [Address] [Data] [N]      VERIFY Fill all N addresses from Address++ with Data.  
--  |    WRITE      [Address] [Data]           Write to address (hex mode generic width) the data.
--  |                                          Z and X and U are also supported.
--  |    FILL       [Address] [Data] [N]       Fill all N addresses from Address with Data.  
--  |
--  |    GLOBAL_RESET                          Resets the ResetN hardware line and the global
--  |                                          reset line for the simulation elements.
--  |   READ_FILE  [Hex Adrs][Inc][File_name] Read from Address, compare to contents of
--  |                                          file as long as the file is not empty, next address
--  |                                is incremented by Inc.
--  |                                Address is Hex accoring to generic address width.
--  |                                Inc is a natural, 0 or larger.
--  |
--  |   WRITE_FILE [Hex Adrs][Inc][file_name] Write to Address, from a file as long as the file is
--  |                                not empty, next address is incremented by Inc.
--  |                                Address is Hex accoring to generic address width.
--  |                                Inc is a natural, 0 or larger.
--  |

-- 
LIBRARY std;
USE std.textio.all;

LIBRARY ieee,bus_tester;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_textio.all;
USE ieee.numeric_std.all;
USE ieee.std_logic_unsigned.all;
USE work.Sital_textio.all;

ENTITY CPU_Simple IS
   GENERIC( 
      command_file   : string  := "CPU_Simple.dat";
      Address_length : integer := 16;
      Data_length    : integer := 16
   );
   PORT( 
      clock     : IN     STD_LOGIC;
      ResetN    : OUT    STD_LOGIC;
      Address   : OUT    std_logic_vector (Address_length-1 DOWNTO 0);
      Data      : INOUT  std_logic_vector (Data_length-1 DOWNTO 0);
      Rd        : OUT    std_logic;
      Wr        : OUT    std_logic;
      Interrupt : IN     STD_LOGIC
   );

-- Declarations

END CPU_Simple ;
-- hds interface_end
-- hds interface_end



ARCHITECTURE Sital_Ofer_Hofman OF CPU_Simple IS

constant MAX_COMMANDS : integer := 2000; 
constant comment_length : integer := 80;
constant cmd_log_string_length : integer :=80;
constant Delay : time := 5 ns;

TYPE CommandType is
    ---standard commands
   (nul, BREAK, WAIT_FOR, LOOP_FOR, END_LOOP, 
    READ, WRITE, GLOBAL_RESET, CONSTANTS, READ_FILE, WRITE_FILE, FILL, VERIFY_FILL,
    Wait_Interrupt);

Type LRM_Type is array (CommandType) of string (1 to 60);

constant LRM : LRM_type := 
(
"                                                            ",
"BREAK                                                       ",
"WAIT_FOR  [T:time]                                          ",
"LOOP_FOR  [I:integer]                                       ",
"END_LOOP                                                    ",
"READ [Hex Adderss] [Hex Data]                               ",
"WRITE [Hex Adderss] [Hex Data]                              ",
"GLOBAL_RESET                                                ",
"CONSTANTS [clk_cycle : time]                                ",
"READ_FILE [Hex Adderss] [Inc] [File_name]                   ",
"WRITE_FILE [Hex Adderss] [Inc] [File_name]                  ",
"FILL [Hex Adderss] [Hex Data] [Repeated]                    ",
"VERIFY_FILL [Hex Adderss] [Hex Data] [Repeated]             ",
"Wait_Interrupt                                              "
);
 

TYPE CommandRecordType IS RECORD
    Command  : CommandType;
    linenum : natural;
    I       : natural; 
    T       : time;
    Addr    : STD_LOGIC_VECTOR(Address_length-1 downto 0);
    Data    : STD_LOGIC_VECTOR(Data_length-1 downto 0);
    file_name : string(1 to cmd_log_string_length);
    ID        : natural;
    slv       : std_logic_vector(31 downto 0);
    comment   : string(1 to comment_length);
	cmd_log   : string(1 to cmd_log_string_length);
END RECORD;

TYPE CommandRecordArrayType IS ARRAY (1 TO MAX_COMMANDS)
    OF CommandRecordType;

TYPE no_time_type is (with_time, without_time);

    ----------------------------
    --Enumerated Read procedure
    ----------------------------
    --This procedure will read from a file enumerated 
    --values who were defined in the CommandType Type.
    --CommandType left value must be nul
    --Line that begin with #, will return values nul and good.
    --White Lines (ENTER) will return values nul and false.
    --spaces in the beginig of a line will be ignored 
    --reading a command who is not defined in the CommandType
    --will return values nul and false
    --
    --Notes :
    --  Case insensitive since enums are always lower case.
    --  empty lines will return True with nul.
    --  comment lines will return true with nul.
    --  
    procedure READ(L:inout LINE; 
                   VALUE:out CommandType;
                   GOOD: out BOOLEAN) IS

     variable c: character;
     variable s: string(1 to 20);
     variable readOk: BOOLEAN;
     variable i : natural;
     variable pos : positive := L'left;

    begin

    --skip white spaces
     Skip_white(L,pos);
     Shrink_line(L,pos);

     good := false;
     value := nul;
     i := 1;

     if l'length = 0 then 
            value := CommandType'Left;
            good := true;
            return;
     end if; 
     
     loop                   
        read(l,c,readOk);   
        if readOk=false or c=' ' or c=HT then
            exit;
        elsif c='#' then
            good := true;
            return;
        else
            --change to lower case since the enumerated type is 
            --always analysed as lower case even if the type is upper case.
            s(i):= To_Lower(c);
            if i=20 then exit; end if;
            i:=i+1;
        end if;
     end loop;

    -- for j in value'base'Left to value'base'Right loop
    --  if VALUE'base'Image(j)=s(1 to i-1) then
     for j in CommandType'Left to CommandType'Right loop
        if CommandType'Image(j)=s(1 to i-1) then
            value :=j;
            good := true;
            return;
        end if;
     end loop;
     return;

     END;


    procedure read_filename(l : inout line; 
                            s : inout string(1 to cmd_log_string_length);
                            good : out boolean) IS
	    VARIABLE pos : positive ;
    begin
        Skip_white(L,pos);
        Shrink_line(L,pos);

        if l'length = 0 then 
            S := (others => ' ');
            good := false;
            return;
        end if; 

        good := true;
        IF l'length > s'length then 
            s := l(l'left to l'left+s'length-1);
        ELSE
            s:=(others=>' ');
            s(1 to l'length) := l.all;
        END IF;
        return;
    end;

    procedure pars_err(cmd : in CommandType;
                       command_file : in string;
                       LineNum : in integer;
					   s : in string;
                       err_flag : inout boolean ) is
        VARIABLE l : line := NULL;

        begin
			if cmd = nul then
				write(l,string'(">> Invalid command in file: "&command_file&",  "));
			else
				write(l,string'(">> Invalid  ' " & CommandType'image(cmd) & " '  command in file: "&command_file&",  "));
			end if;
			writeline(output,l);
			write(l,string'(">> Line #"& integer'image(LineNum)&" = " & s));
			writeline(output,l);
			
			if cmd /= nul then
				write(l,string'(">> Valid format: " & LRM(cmd)));
				writeline(output,l);
			end if;
			write(l,string'(">>"));
			writeline(output,l);
			err_flag := true;   
	end;


    PROCEDURE parser ( command_file : IN string;
                       commands : INOUT CommandRecordArrayType;
                       NumCommands : OUT Natural;
                       err_flag : INOUT boolean   ) IS
            
        FILE TestFile : text;
        VARIABLE l : line := NULL;
        VARIABLE linenum,PC : NATURAL := 0;
        VARIABLE EndOfFile,good : BOOLEAN := FALSE;
        VARIABLE int : integer;
        VARIABLE tm : TIME;
        VARIABLE cmnd : CommandType;
        VARIABLE s : string(1 to cmd_log_string_length);
        VARIABLE fname : string(1 to 80);
	    VARIABLE pos : positive ;
        VARIABLE id : natural;
        VARIABLE slv : std_logic_vector(31 downto 0);

    BEGIN
       FILE_OPEN( TestFile, command_file);
       all_file : LOOP

         Readline(TestFile,l);

		 pos := L'left;
		 Skip_white(L,pos);
		 Shrink_line(L,pos);

         s := (others=>' ');
         IF l'length > cmd_log_string_length then 
            s := l(l'left to l'left+(cmd_log_string_length-1));
         ELSE
            s(1 to  l'length):= l.all(l'range);
         END IF;

         lineNum := lineNum +1;
         EndOfFile := endfile(TestFile);

         cmnd := nul;
         read(l,cmnd,good);

         exit when not good;

         PC := PC +1;
         commands(PC).command := cmnd;
         commands(PC).lineNum := lineNum;
		 commands(PC).cmd_log := s;

         CASE cmnd IS

            WHEN WAIT_FOR  => 
                read(l,commands(PC).T, good);
                exit when NOT good;

            WHEN LOOP_FOR => 
                read(l,commands(PC).I, good);
                exit when NOT good;

            WHEN BREAK | END_LOOP | GLOBAL_RESET | Wait_Interrupt=> 
                null;

           WHEN READ  => 
                hRead(l,commands(PC).addr,good);
                exit when NOT good;
                ReadHex(l,commands(PC).data,good);
                exit when NOT good;

            WHEN WRITE  => 
                hRead(l,commands(PC).addr,good);
                exit when NOT good;
                ReadHex(l,commands(PC).data,good);
                exit when NOT good;

            WHEN FILL | VERIFY_FILL => 
                hRead(l,commands(PC).addr,good);
                exit when NOT good;
                ReadHex(l,commands(PC).data,good);
                exit when NOT good;
                read(l,commands(PC).I, good);
                exit when NOT good;

            WHEN READ_FILE | WRITE_FILE => 
                HRead(l,commands(PC).addr,good);
                exit when NOT good;
                Read(l,commands(PC).I,good);
                exit when NOT good;
                read_filename(l,commands(PC).file_name,good);
                exit when NOT good;

            WHEN nul => --empty line or comment line.
                PC := PC -1;
               
            WHEN OTHERS => 
               good := false;
               exit;       

         END CASE;
        
         NumCommands := PC;
         EXIT WHEN EndOfFile;
       END LOOP;

       FILE_CLOSE( TestFile );

       IF NOT good THEN
       	 pars_err(cmnd, command_file, LineNum, s, err_flag); 
         report " " severity failure;
       END IF;

    END parser;


    SIGNAL clock_cycle  : time := 20 ns;
    SIGNAL clock_int  : STD_LOGIC := '0';
    SIGNAL bfm_command: CommandType;
    SIGNAL N : natural;

BEGIN

    clock_int <= clock;

 PROCESS
     File       Fp : text;
     VARIABLE   l : line;
     VARIABLE   commands : CommandRecordArrayType;
     VARIABLE   NumCommands : Natural :=0;
     VARIABLE   err_flag : boolean := false;
     VARIABLE   pc : natural;
     VARIABLE   loop_back_location : natural;
     VARIABLE   loop_counter : natural;
     VARIABLE   Lp : line;
     VARIABLE   FData : STD_LOGIC_VECTOR(Data_length-1 downto 0);
     VARIABLE   FAddr : STD_LOGIC_VECTOR(Address_length-1 downto 0);
     VARIABLE   Good : boolean;
     VARIABLE   Int : natural;

----------------------------------------------------
	  Procedure Read_cycle(
                             Ad  : STD_LOGIC_VECTOR(Address_length-1 downto 0);
                             expected_Data : inout STD_LOGIC_VECTOR(Data_length-1 downto 0)
                          ) IS
      BEGIN
          Address <= Ad;
    	  data <= (others=>'Z');
          wait until clock_int'event and clock_int = '1';
          Rd <= '1' after delay;
  		  wait until clock_int'event and clock_int = '1';
  		  wait until clock_int'event and clock_int = '1';
  		  wait until clock_int'event and clock_int = '1';
  		  Rd <= '0'  after delay;
          -- Examine Data Value     
          IF not std_match(Data,expected_data) then
  			report "**** CPU_BFM Warning - READ Violation;  Expected Data= "&
  			     vec2hexstring(expected_data)&
  			     " , Actual Data Value= "&
  			     vec2hexstring(data) & " On line number " & 
  			     integer'image(commands(pc).linenum) severity error;
          end if;
          expected_Data := Data;
  		  wait until clock_int'event and clock_int = '1';
  		  Address <= (OTHERS=>'1') after delay;
          wait until clock_int'event and clock_int = '1';
       END;

----------------------------------------------------
	  Procedure Write_cycle(
                             Ad  : STD_LOGIC_VECTOR(Address_length-1 downto 0);
                             D   : STD_LOGIC_VECTOR(Data_length-1 downto 0)
                          ) IS
      BEGIN
          Address <= Ad  after delay;
    	  data <= D  after delay;
          wait until clock_int'event and clock_int = '1';
  		  Wr <= '1'  after delay;
          wait until clock_int'event and clock_int = '1';
          Wr <= '0'  after delay;
  		  wait until clock_int'event and clock_int = '1';
  		  Address <= (OTHERS=>'1') after delay;
          Data <= (OTHERS=>'Z') after delay;
          wait until clock_int'event and clock_int = '1';
       END;

----------------------------------------------------
	  Procedure Fill_cycle(
                             Ad  : STD_LOGIC_VECTOR(Address_length-1 downto 0);
                             D   : STD_LOGIC_VECTOR(Data_length-1 downto 0);
                             L   : in Integer
                          ) IS
      BEGIN
      	  data <= D  after delay;
          For i in 0 to L-1 loop
              Address <= Ad+i after delay;             
      		  Wr <= '1'  after delay;
              wait until clock_int'event and clock_int = '1';
          end loop;
          Wr <= '0'  after delay;
  		  Address <= (OTHERS=>'1') after delay;
          Data <= (OTHERS=>'Z') after delay;
          wait until clock_int'event and clock_int = '1';
       END;

----------------------------------------------------
	  Procedure Verify_Fill_cycle(
                             Ad  : STD_LOGIC_VECTOR(Address_length-1 downto 0);
                             expected_data   : STD_LOGIC_VECTOR(Data_length-1 downto 0);
                             L   : in Integer
                          ) IS
      BEGIN
          wait until clock_int'event and clock_int = '1';
          Data <= (OTHERS=>'Z') after delay;
          Address <= Ad after delay;             
  		  Rd <= '1'  after delay;
          wait until clock_int'event and clock_int = '1';
          
          For i in 1 to L-1 loop
                  Address <= Ad+i after delay;             
          		  Rd <= '1'  after delay;
                  wait until clock_int'event and clock_int = '1';
                  IF not std_match(Data,expected_data) then
          			report "**** CPU_BFM Warning - VERIFY FILL Violation;  Expected  Address "&
          			     vec2hexstring(Ad+i-1)& "  Expected Data= "&
          			     vec2hexstring(expected_data)&
          			     " , Actual Data Value= "&
          			     vec2hexstring(data) & " On line number " & 
          			     integer'image(commands(pc).linenum) severity error;
                  end if;
          end loop;
          ---now verify last word
          wait until clock_int'event and clock_int = '1';
          IF not std_match(Data,expected_data) then
  			report "**** CPU_BFM Warning - VERIFY FILL Violation;  Expected  Address "&
  			     vec2hexstring(Ad+L-1)& "  Expected Data= "&
  			     vec2hexstring(expected_data)&
  			     " , Actual Data Value= "&
  			     vec2hexstring(data) & " On line number " & 
  			     integer'image(commands(pc).linenum) severity error;
          end if;
          Rd <= '0'  after delay;
  		  Address <= (OTHERS=>'1') after delay;
          wait until clock_int'event and clock_int = '1';
       END;

----------------------------------------------------


  BEGIN

    data <= (others=>'Z');
    address <= (others=>'1');
    Rd <= '0';
    Wr <='0';

    --clear screen 
    for i in 1 to 10 loop
        write(l,string'(">>")); writeline(output,l);
    end loop;

    parser (command_file, commands, NumCommands, err_flag);

    execution_loop :  LOOP

         pc:=pc+1; -- Program Counter  
         bfm_command <= Commands(pc).command; --for debug;

         CASE Commands(pc).command IS

            WHEN LOOP_FOR => 
                LOOP_Counter := Commands(pc).I;
                loop_back_location := PC;

            WHEN END_LOOP => 
                IF (loop_counter > 1) THEN
                  PC := loop_back_location;
                  Loop_counter := Loop_counter -1;    
                END IF;

            WHEN WAIT_FOR => 
                wait for commands(PC).T;
                wait until clock_int'event and clock_int = '1';

            WHEN BREAK =>
                report " (not a failure) => CPU simple has reached a BREAK command and stopped execution, thank you and good bye !!"
                SEVERITY failure;

            WHEN nul => ---nul would be the first location in the command array that was 
                        ---    not parsed, i.e. end of commands.
                Exit execution_loop;          

            WHEN GLOBAL_RESET  => 
                ResetN <= '0', '1' after 100 ns;

            WHEN READ  =>  
				Read_cycle(commands(pc).addr, commands(pc).data);

            WHEN VERIFY_FILL  =>  ---burst read
                Verify_Fill_cycle(commands(pc).addr, commands(pc).data, commands(pc).I);
				
            WHEN WRITE  =>  
				write_cycle(commands(pc).addr, commands(pc).data);
				
            WHEN FILL  =>  ---burst write
                Fill_cycle(commands(pc).addr, commands(pc).data, commands(pc).I);
				
            WHEN READ_FILE => 
                Int := 1;
                FAddr := commands(PC).addr;
                FILE_OPEN( Fp, commands(PC).file_name, read_mode);
                While not endfile(Fp) Loop
                    readline(Fp, Lp);
                    ReadHex(Lp, FData, Good);
                    IF not Good THEN
                      FILE_CLOSE( Fp );
                      REPORT "CPU_BFM : reading data from file "& commands(PC).file_name & cr & 
                             " for READ_FILE command was not successful. " & 
                             "Please check if data is in the correct format and width. " & 
                             "Expected format is a column if Hex nibbles. " & CR & 
                             "Error in "& command_file &" in line number " & integer'image(commands(pc).linenum) &
                             " Error was found when reading the " & integer'image(Int) &
                             " elment of the file. ...Stopping simulation."
                      SEVERITY Failure;
					ENd IF;
                    read_cycle(FAddr, Fdata);
                    Int := Int + 1;
                    FAddr := FAddr + commands(PC).I;
                end loop;
                FILE_CLOSE( Fp );
                
            WHEN WRITE_FILE => 
                Int := 1;
                FAddr := commands(PC).addr;
                FILE_OPEN( Fp, commands(PC).file_name, read_mode);
                While not endfile(Fp) Loop
                    readline(Fp, Lp);
                    ReadHex(Lp, FData, Good);
                    IF not Good THEN
                      FILE_CLOSE( Fp );
                      REPORT "CPU_BFM : reading data from "& commands(PC).file_name & cr & " for WRITE_FILE command was not successful. " & 
                             "Please check if data is in the correct format and width. " & 
                             "Expected format is a column if Hex nibbles. " & CR & 
                             "Error in "& command_file &" in line number " & integer'image(commands(pc).linenum) &
                             " Error was found when reading the " & integer'image(Int) &
                             " elment of the file. ...Stopping simulation."
                      SEVERITY Failure;
					ENd IF;
                    write_cycle(FAddr, Fdata);
                    Int := Int + 1;
                    FAddr := FAddr + commands(PC).I;
                end loop;
                FILE_CLOSE( Fp );
				 
            WHEN Wait_Interrupt => 
                wait until interrupt'event and interrupt = '1';
         
            WHEN OTHERS => null;

            END CASE;
      END LOOP execution_loop;

	  report cr&"**** End Of All Tests --- BFM Simulation Halted ****"&cr&cr
	         SEVERITY failure;
      wait;
END PROCESS;
END Sital_Ofer_Hofman;


