---
---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---
---
---  This file is limited for use by customers of Sital Technology only.
---
---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---
---
---
---  The executer_simple is designed to execute MIL-STD-1553 communications
---   from the tester Mux bus side.
---  The bus_tester controls the receivers and transmitters for both channel
---   A and B.
---  The executer_simple reads data from a the bus_tester.DAT file.
---
---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---
---
---          Designed by Ofer Hofman from Sital Technology.
---           Initail version Aug 2004.
---
---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---
-- -
-- 
--  Language supported constructs :
--
--
--  <word>  is a 4 nibble element such as AB01.
--  
--  <1553 word> could be on of 
--  
--      "C" <TA T/R SA WC> or                Command sync with command data built from Terminal Address, 
--                                           Transmit or Receive, Sub Address, and Word Count all in Hex.
--      "S" <word> or                        Command sync with word
--      "D" <word>                           data sync with word
--  
--  <message> could be one or more <1553 word> elements. 
--  
--  
--      example:  C 18 R 01 03      D 2222   D 3333    D 1234     S C000
--  
--  
--  <message status> could be of the form :
--      CS                                   Clear Status
--      NR                                   No Response
--      
--  Language supported commands :
--  
--  TxA  <message>
--  TxB  <message>
--  
--  WAIT_UNTIL  RxA or RxB or  TxA_Rdy  or TxB_Rdy   Wait until the Receiver A or B or transmitter A or B are ready.
--                      example : WAIT_UNTIL RxA   will wait until a word was received from bus A.
--  
--  wait_for  <time>                        WAIT for time delay. Time delay could be 1 ns or 23 us or 3456 ns ...
--                                          no decimal point in time unit.
--  
--  checkA  <message status>                Checks Either Clear Status or No Response.
--  CheckB  <message status>                Checks Either Clear Status or No Response.
--   
--  echo                                    40 characters echoed to screen
--  break                                   stops the program.
--  ReceiveA  <message>                     receives a message on bus A.
--  ReceiveB  <message>                     receives a message on bus B.
-- 
--  Pulse     <time> <value>                inserts a pulse to the bus. Width is <time> and value is 1 or 0 for '1' or '0'
--
--
LIBRARY ieee ;
USE std.textio.all;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;
USE std.textio.all;
USE ieee.std_logic_textio.all;
USE work.Sital_textio.all;

ENTITY Executer_simple2 IS
   GENERIC( 
      comm_file_name : string := "../hds2004/bus_tester.dat"
   );
   PORT( 
      -- transmitter_A INterface --
      TA_Go           : OUT    boolean;
      TA_Sync         : OUT    std_logic;
      TA_Data         : OUT    STD_LOGIC_VECTOR (15 DOWNTO 0);  --MSB sent first.
      TA_Transmitting : IN     boolean;
      TA_Pulse        : out    std_logic;
      -- transmitter_A INterface --
      TB_Go           : OUT    boolean;
      TB_Sync         : OUT    std_logic;
      TB_Data         : OUT    STD_LOGIC_VECTOR (15 DOWNTO 0);  --MSB sent first.
      TB_Transmitting : IN     boolean;
      TB_Pulse        : out    std_logic;
      -- Receiver A --
      RA_receiveing   : IN     BOOLEAN;
      RA_RxRdy        : IN     BOOLEAN;
      RA_Sync         : IN     STD_LOGIC  := '0';
      RA_Data         : IN     STD_LOGIC_VECTOR (15 DOWNTO 0);  --MSB sent first.
      RA_CS           : IN     boolean;
      RA_Gap          : IN     time       := 4 us;
      -- Receiver B --
      RB_receiveing   : IN     BOOLEAN;
      RB_RxRdy        : IN     BOOLEAN;
      RB_Sync         : IN     STD_LOGIC  := '0';
      RB_Data         : IN     STD_LOGIC_VECTOR (15 DOWNTO 0);  --MSB sent first.
      RB_CS           : IN     boolean;
      RB_Gap          : IN     time       := 4 us
   );

-- Declarations

END Executer_simple2 ;


ARCHITECTURE SITAL_TECHNOLOGY OF Executer_simple2 IS

	TYPE command_type   IS 
--commands
	    (NUL, TxA, TxB, wait_until, wait_for, checkA, CheckB, 
	     echo, break, ReceiveA, ReceiveB, PulseA, PulseB,
--parameters for commands
         C, D, S, T, R, RxA, RxB, TxA_Rdy, TxB_Rdy, CS, NR, ME,  
         reportA, reportB, Time_out );

    type message_type is array (1 to 40) of std_logic_vector(16 downto 0);

	TYPE CommandRecordType IS RECORD
	   Command 	 		: COMMAND_TYPE;
       line_no          : natural;
       messege_block    : message_type;
       messege_length   : natural;
       param1           : COMMAND_TYPE;
       int1             : natural;
       t                : delay_length;
	   comment          : string(1 to 60);
	END RECORD;

	CONSTANT MAX_COMMANDS : natural := 1000;
	TYPE CommandRecordArrayType IS ARRAY (1 TO MAX_COMMANDS) OF CommandRecordType;

    procedure READ(L:inout LINE; 
                   VALUE:out command_type;
                   GOOD: out BOOLEAN) IS

     variable c: character;
     variable s: string(1 to 20);
     variable readOk: BOOLEAN;
     variable i : natural;
     variable pos : positive := L'left;

    begin

    --skip white spaces
     Skip_white(L,pos);
     Shrink_line(L,pos);

     good := false;
     value := nul;
     i := 1;

     if l'length = 0 then 
    		good := true;
    		return;
     end if; 
     
     loop					
     	read(l,c,readOk);	
    	if readOk=false or c=' ' or c=HT then
    		exit;
    	elsif c='#' then
    		good := true;
    		return;
    	else
    		--chage to lower case since the enumerated type is 
    		--always analysed as lower case even if the type is upper case.
    		s(i):= To_Lower(c);
    		i:=i+1;
    	end if;
     end loop;

    -- for j in value'base'Left to value'base'Right loop
    --	if VALUE'base'Image(j)=s(1 to i-1) then
     for j in command_type'Left to command_type'Right loop
    	if command_type'Image(j)=s(1 to i-1) then
    		value :=j;
    		good := true;
    		return;
    	end if;
     end loop;
     return;

     END;


	PROCEDURE parser ( comm_file_name    : IN string;
	                   commands       : INOUT  CommandRecordArrayType;
	                   NumCommands       : OUT Natural;
	                   err_flag       : INOUT boolean ) IS
	            
	            FILE command_file : text ; -- command file

	            VARIABLE l              : line := NULL;
                VARIABLE ct,p1          : COMMAND_TYPE;
	            VARIABLE linenum,CmdNum : NATURAL := 0;
	            VARIABLE EndOfFile,good : BOOLEAN := FALSE;
	            VARIABLE good_param     : BOOLEAN := FALSE;
                VARIABLE status : file_open_status;

                procedure get_message(L : inout Line; 
                          messege_block : out message_type;
                          messege_length : out natural; 
                          param : out COMMAND_TYPE;
                          good : out boolean) IS
                    VARIABLE byte    : Std_logic_vector(7 downto 0);
	                VARIABLE very_good  : BOOLEAN := FALSE;
                    VARIABLE v1, vp, p1 : COMMAND_TYPE;--vp :: v1 previous
                    VARIABLE Length     : natural;
                          
                begin
                    length := 0; 
                    param := nul;
                    messege_length := 0;
                    messege_block := (others => (others => '0'));
                    vars : for i in messege_block'range loop
                        vp := v1;
                        read(l, v1, very_good);
                        IF not (very_good) THEN 
                            good := false; 
                            return; 
                        ELSE --could be a empty till end of line or comment.
                        case v1 is
                            WHEN C =>  messege_block(i)(16) := '1';
                                       --read address
                                       ReadHex(l,byte,very_good);
                                       IF not (very_good) THEN good := false; return; END IF;
                                       messege_block(i)(15 downto 11) := byte(4 downto 0);
                                       --read T or R
                                       read(l, p1, very_good);
                                       IF not (very_good) THEN good := false; return; END IF;
                                       CASE (p1) IS
                                        WHEN T => messege_block(i)(10) := '1';
                                        WHEN R => messege_block(i)(10) := '0';
                                        WHEN OTHERS => good := false; return;
                                       END CASE;  
                                       IF not (very_good) THEN good := false; return; END IF;
                                       --read Subaddress
                                       ReadHex(l,byte,very_good);
                                       IF not (very_good) THEN good := false; return; END IF;
                                       messege_block(i)(9 downto 5) := byte(4 downto 0);
                                       --read word count
                                       ReadHex(l,byte,very_good);
                                       IF not (very_good) THEN good := false; return; END IF;
                                       messege_block(i)(4 downto 0) := byte(4 downto 0);

                            WHEN S =>  messege_block(i)(16) := '1';
                                       ReadHex(l,messege_block(i)(15 downto 0),very_good); 
                            WHEN D =>  messege_block(i)(16) := '0';
                                       ReadHex(l,messege_block(i)(15 downto 0),very_good); 
                            WHEN CS => param := CS; 
                            WHEN NR => param := NR;
                            WHEN ME => param := ME;
                            WHEN nul=> if i = 1 then --comment or empty line
                                            good := false;
                                            return;
                                       else 
                                            good := true;
                                            exit vars;
                                       end if;
                            WHEN OTHERS => report "One of the message words was in a wrong format, please correct and try again " &
                                                  "The line number of the error will be on the next warning"
                                       severity error;
                                       very_good := false;
                        end case;
                        end if;
                        IF not (very_good) THEN good := false; return; END IF;
                      length := i;
                      --messege_length := i;
                    end loop;
                    IF length = 2 and (vp = ME or vp = NR or vp = CS) then 
                        messege_length := 1;
                    ELSE
                        messege_length := length;
                    end if;
                    return;
                end get_message;

	BEGIN
	  file_open(status, command_file,comm_file_name);
        IF (status /= open_ok) THEN
            report "Opening file name : " & comm_file_name & 
                    " resulted with the following error : " & 
                    file_open_status'image(status)
            severity failure;
        END IF;
	  parse : LOOP
    	Readline(command_file,l);
    	EndOfFile := endfile(command_file);
        CmdNum  := CmdNum  + 1;
        lineNum := lineNum + 1;
        read(l, ct, good); 
        if not good then --was not a valid command or comment or empty line.
            report "On line: "&integer'image(lineNum) &" You wrote a line which was not "& 
                   "a valid command nor a comment nor an empty line. Please correct and rerun test"
            severity failure;
            exit parse;
        else --was a valid command or comment or empty line.
	      commands(CmdNum).command := ct; --common for all commands
          commands(CmdNum).line_no := lineNum;
	      CASE ct IS
            WHEN NUL =>         --was a comment or empty line.
                                CmdNum := CmdNum -1;
            WHEN TxA | TxB =>   get_message(l,commands(CmdNum).messege_block,
                                              commands(CmdNum).messege_length,
                                              commands(CmdNum).param1,good_param);
                                exit parse when not good_param;

            WHEN wait_until =>  read(l,commands(CmdNum).param1,good_param);
                                exit parse when not good_param;
                                IF (commands(CmdNum).param1 /= RxA and 
                                    commands(CmdNum).param1 /= RxB and 
                                    commands(CmdNum).param1 /= TxA_Rdy and 
                                    commands(CmdNum).param1 /= TxB_Rdy  )
                                    THEN
                                        good_param := false;
                                        report "WAIT_UNTIL expected RxA or RxB or TxA_Rdy or TxB_Rdy but found " & 
                                                command_type'image(commands(CmdNum).param1) & 
                                               " The line number of the error will be on the next warning";
                                END IF;
                                exit parse when not good_param;
            WHEN wait_for =>    read(l,commands(CmdNum).t,good_param);
                                exit parse when not good_param;

            WHEN PulseA | PulseB =>    
                                read(l,commands(CmdNum).t,good_param);
                                exit parse when not good_param;
                                read(l,commands(CmdNum).int1,good_param);
                                exit parse when not good_param;
                                good_param := (commands(CmdNum).int1 = 1) or (commands(CmdNum).int1 = 0);
                                IF not good_param THEN
                                    Report "Pulse command 2nd parameter should be 1 or 0, but it is " & integer'image(commands(CmdNum).int1)
                                    severity warning;
                                    exit parse;
                                ENd IF;

            WHEN checkA | CheckB =>
                                get_message(l,commands(CmdNum).messege_block,
                                              commands(CmdNum).messege_length,
                                              commands(CmdNum).param1,good_param);
                                exit parse when not good_param;
                                if not (commands(CmdNum).messege_length = 1 or
                                    commands(CmdNum).param1 = CS or commands(CmdNum).param1 = NR) then 
                                    report COMMAND_TYPE'image(ct) & " command in bus_tester accepts one message parameter but got " & 
                                                                    integer'image(commands(CmdNum).messege_length) & 
                                                                    " parameters. Line number of Error will be in the next Note"
                                    severity error;
                                    good_param := false;
                                    exit parse;
                                end if;
                                exit parse when not good_param;
            WHEN ReceiveA | ReceiveB =>
                                get_message(l,commands(CmdNum).messege_block,
                                              commands(CmdNum).messege_length,
                                              commands(CmdNum).param1,good_param);
                                exit parse when not good_param;
                                if commands(CmdNum).messege_length < 1  then 
                                    report COMMAND_TYPE'image(ct) & " command in bus_tester accepts one or more message parameter but got 0 " & 
                                                                    " parameters. Line number of Error will be in the next Note"
                                    severity error;
                                    good_param := false;
                                    exit parse;
                                end if;
                                exit parse when not good_param;
			WHEN ECHO | BREAK =>ReadString(l,commands(CmdNum).comment);

	        WHEN OTHERS =>
	                    REPORT ">> Invalid command in file: "& comm_file_name&",  line #"& integer'image(LineNum) &",  "& 
	                           "Unknown command : " & command_type'image(ct)
	                    SEVERITY error;
	                    err_flag := true;   
                        exit parse;
	      END CASE;
        end if; --good command
	    NumCommands    := CmdNum;
	    EXIT WHEN EndOfFile;
	  END LOOP;
      IF not good_param then 
        report " BAD command/parameter at line :" & integer'image(LineNum) & 
                " Still left for parsing in this line : " & l(l'left to l'right);
	    err_flag := true;   
      end if;
	  file_close(command_file);
	END parser;

    SIGNAL cpu_command 		: CommandRecordType;
    
    SHARED VARIABLE  rd_data		: STD_LOGIC_VECTOR(15 DOWNTO 0)  := (OTHERS=> 'Z');
    
    SIGNAL  CleanStatusA : boolean;
    SIGNAL  CleanStatusB : boolean;

BEGIN --of architecture

    CleanStatusA <= (RA_data(10 downto 0) = "00000000000") and RA_CS;
    CleanStatusB <= (RB_data(10 downto 0) = "00000000000") and RB_CS;

   main: PROCESS
            VARIABLE    commands       : CommandRecordArrayType;
            VARIABLE    NumCommands    : Natural :=0;
            VARIABLE    err_flag,good,EndOfFile   : boolean := false;
            VARIABLE    pc              : natural := 1;
            VARIABLE    l               : line := NULL;
            VARIABLE    WC              : natural;
            VARIABLE    v1              : COMMAND_TYPE;

 BEGIN
   TA_Pulse <= 'Z'; 
   TB_Pulse <= 'Z'; 
   parser (comm_file_name,commands,NumCommands,err_flag);
   IF err_flag THEN
      REPORT " ## !!!!!  ERRORs FOUND IN THE COMMAND FILE - HALTING EXECUTE !!!!! ## "
      SEVERITY failure;
      WAIT;
   ELSE
      execution_loop :  LOOP
        cpu_command <= Commands(pc); --for debug;
        CASE Commands(pc).command IS
            WHEN nul =>      --Only at last command
                           Exit;          
            WHEN  wait_until  => 
                            CASE (Commands(pc).param1) IS
                               WHEN RxA => Wait until RA_RxRdy'event and RA_RxRdy for 50 us;
                                    ASSERT RA_RxRdy report "Wait until data ready FAILED ! receiver didn't get any transmission back as expected. line number "&integer'image(commands(PC).line_no)
                                        severity error;
                               WHEN RxB => Wait until RB_RxRdy'event and RB_RxRdy for 50 us;
                                    ASSERT RB_RxRdy report "Wait until data ready FAILED ! receiver didn't get any transmission back as expected. line number "&integer'image(commands(PC).line_no)
                                        severity error;
                               WHEN TxA_Rdy => Wait until TA_Transmitting'event and (not TA_Transmitting) ;
                               WHEN TxB_Rdy => Wait until TB_Transmitting'event and (not TB_Transmitting) ;
                               WHEN OTHERS => null;
                            END CASE;  
            WHEN wait_for=> WAIT FOR Commands(pc).t;
            
            WHEN PulseA   => IF commands(PC).int1=1 THEN
                                TA_Pulse <= '1';
                             ELSE
                                TA_Pulse <= '0';
                             END If;
                             WAIT FOR commands(PC).t;
                             TA_Pulse <= 'Z';
            
            WHEN PulseB   => IF commands(PC).int1=1 THEN
                                TB_Pulse <= '1';
                             ELSE
                                TB_Pulse <= '0';
                             END If;
                             WAIT FOR commands(PC).t;
                             TB_Pulse <= 'Z';
            

            WHEN TxA  =>    FOR i IN 1 to Commands(pc).messege_length LOOP
                                TA_Data <= Commands(pc).messege_block(i)(15 downto 0);
                                TA_Sync <= Commands(pc).messege_block(i)(16);
                                wait for 1 ns; --create a delta
                                TA_Go <= true, false after 1 ns;
                                wait until TA_Transmitting'event and not TA_Transmitting;
                            END LOOP; 
            WHEN TxB  =>    FOR i IN 1 to Commands(pc).messege_length LOOP
                                TB_Data <= Commands(pc).messege_block(i)(15 downto 0);
                                TB_Sync <= Commands(pc).messege_block(i)(16);
                                wait for 1 ns; --create a delta
                                TB_Go <= true, false after 1 ns;
                                wait until TB_Transmitting'event and not TB_Transmitting;
                            END LOOP; 
            WHEN ReceiveA =>FOR i IN 1 to Commands(pc).messege_length LOOP
                                WAIT UNTIL RA_RxRdy'event and RA_RxRdy;  --no timeout !!
                                ASSERT commands(pc).messege_block(i)(16) = RA_sync
                                   report "ReceiveA : Sync type received on bus A from UUT is not as expected.command at line " & 
                                      integer'image(Commands(pc).line_no) severity error;
                                ASSERT commands(pc).messege_block(i)(15 downto 0) = RA_data
                                   report "ReceiveA : Data received on bus A from UUT is not as expected. command at line " & 
                                      integer'image(Commands(pc).line_no) severity error;
                            END LOOP;  
            WHEN ReceiveB =>FOR i IN 1 to Commands(pc).messege_length LOOP
                                WAIT UNTIL RB_RxRdy'event and RB_RxRdy; --no timeout !!
                                ASSERT commands(pc).messege_block(i)(16) = RB_sync
                                   report "ReceiveB : Sync type received on bus B from UUT is not as expected.command at line " & 
                                      integer'image(Commands(pc).line_no) severity error;
                                ASSERT commands(pc).messege_block(i)(15 downto 0) = RB_data
                                   report "ReceiveB : Data received on bus B from UUT is not as expected. command at line " & 
                                      integer'image(Commands(pc).line_no) severity error;
                            END LOOP;  
            WHEN checkA  => CASE (commands(pc).param1) IS
                            WHEN CS  =>  IF not CleanStatusA THEN 
                                            report "Check bus A is not Clear Ststus (CS) as was expected in the command CheckA CS On line" & 
                                                        integer'image(Commands(pc).line_no) & 
                                   "  Data received is : " & vec2hexstring(RA_data)
                                   severity error;
                                         END IF;
                            WHEN ME  =>  IF (RA_data(10 downto 0) /= "10000000000" or (not RA_CS)) THEN 
                                            report "Check bus A is not ME as was expected in the command CheckA ME On line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                                         END IF;
                            WHEN NR  =>  Wait until RA_RxRdy'event and RA_RxRdy for 50 us;
                                         ASSERT not RA_RxRdy 
                                            report "A No Response was expected but a word was received on bus A. command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                            WHEN nul =>  assert commands(pc).messege_block(1)(16) = RA_sync
                                            report "Sync type received on bus A from UUT is not as expected.command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                                         ASSERT commands(pc).messege_block(1)(15 downto 0) = RA_data
                                            report "Data received on bus A from UUT is not as expected. command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                            WHEN others => report "CHECKA condition not supported";
                            END CASE;  
            WHEN CheckB  => CASE (commands(pc).param1) IS
                            WHEN CS  =>  IF not CleanStatusB THEN 
                                            report "Check bus B is not Clear Ststus (CS) as was expected in the command CheckB CS On line "& 
                                                        integer'image(Commands(pc).line_no) & 
                                           "  Data received is : " & vec2hexstring(RA_data)
                                           severity error;
                                         END IF;
                            WHEN ME  =>  IF (RB_data(10 downto 0) /= "10000000000" or (not RB_CS)) THEN 
                                            report "Check bus B is ME as was expected in the command CheckB ME On line "& 
                                                        integer'image(Commands(pc).line_no) severity error;
                                         END IF;
                            WHEN NR  =>  Wait until RB_RxRdy'event and RB_RxRdy for 50 us;
                                         ASSERT not RB_RxRdy 
                                            report "A No Response was expected but a word was received on bus B. command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                            WHEN nul =>  assert commands(pc).messege_block(1)(16) = RB_sync
                                            report "Sync type received on bus B from UUT is not as expected. command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                                         ASSERT commands(pc).messege_block(1)(15 downto 0) = RB_data
                                            report "Data received on bus B from UUT is not as expected. command at line " & 
                                                        integer'image(Commands(pc).line_no) severity error;
                            WHEN others => report "CHECKB condition not supported";
                            END CASE;  
			WHEN ECHO     =>report cr & ht & ht & commands(pc).comment;
	        WHEN break    =>report " (not a failure) => bus_tester has reached a BREAK command and stopped execution, thank you and good bye !!"
                            severity failure;
            WHEN OTHERS =>  NULL;
          END CASE;
          pc:=pc+1;
        END LOOP execution_loop;
    WAIT;
  END IF; --errors in parser;
  wait;
END PROCESS;
   
    
	
END sital_technology;                    



